"""
.. include:: ../README.md
.. include:: ../CHANGELOG.md
.. include:: ../pdocs/user_guide/1_concepts.md
.. include:: ../pdocs/user_guide/2_credentials.md
.. include:: ../pdocs/user_guide/3_postprocess.md
.. include:: ../pdocs/user_guide/4_pipelines.md
.. include:: ../pdocs/user_guide/5_data_management.md
.. include:: ../pdocs/developer_guide/1_main.md
"""



from typing import Any, Dict, List, Optional, Union

from landingai.data_management.client import METADATA_GET, METADATA_UPDATE, LandingLens
from landingai.data_management.utils import (
    PrettyPrintable,
    obj_to_dict,
    metadata_to_ids,
    ids_to_metadata,
)


class Metadata:
    """Metadata management API client.
    This class provides a set of APIs to manage the metadata of the medias (images) uploaded to LandingLens.
    For example, you can use this class to update the metadata of the uploaded medias.

    Example
    -------
    >>> client = Metadata(project_id, api_key)
    >>> client.update([101, 102, 103], creator="tom")

    Parameters
    ----------
    project_id: int
        LandingLens project id.  Can override this default in individual commands.
    api_key: Optional[str]
        LandingLens API Key. If it's not provided, it will be read from the environment variable LANDINGAI_API_KEY, or from .env file on your project root directory.
    """

    def __init__(self, project_id: int, api_key: Optional[str] = None):
        self._client = LandingLens(project_id=project_id, api_key=api_key)

    def update(
        self,
        media_ids: Union[int, List[int]],
        **input_metadata: Optional[Dict[str, Any]],
    ) -> Dict[str, Any]:
        """Update or insert a dictionary of metadata for a set of medias.

        Parameters
        ----------
        media_ids
            Media ids to update.
        input_metadata
            A dictionary of metadata to be updated or inserted. The key of the metadata
            needs to be created/registered (for the first time) on LandingLens before
            calling update().

        Returns
        ----------
        Dict[str, Any]
            The result from the update().
            ```
            # Example output
            {
                "project_id": 12345,
                "metadata": [...],
                "media_ids": [123, 124]],
            }
            ```
        """
        project_id = self._client._project_id
        if (
            not media_ids
            or isinstance(media_ids, bool)
            or (not isinstance(media_ids, int) and len(media_ids) == 0)
        ):
            raise ValueError("Missing required flags: {'media_ids'}")

        if not input_metadata or len(input_metadata) == 0:
            raise ValueError("Missing required flags: {'metadata'}")

        dataset_id = self._client.get_project_property(project_id, "dataset_id")

        if isinstance(media_ids, int):
            media_ids = [media_ids]
        else:
            # to avoid errors due to things like numpy.int
            media_ids = list(map(int, media_ids))

        metadata_mapping, id_to_metadata = self._client.get_metadata_mappings(
            project_id
        )

        body = _MetadataUploadRequestBody(
            selectOption=_SelectOption(media_ids),
            project=_Project(project_id, dataset_id),
            metadata=metadata_to_ids(input_metadata, metadata_mapping),
        )

        resp = self._client._api(METADATA_UPDATE, data=obj_to_dict(body))
        resp_data = resp["data"]
        return {
            "project_id": project_id,
            "metadata": ids_to_metadata(resp_data[0]["metadata"], id_to_metadata),
            "media_ids": [media["objectId"] for media in resp_data],
        }

    def get(self, media_id: int) -> Dict[str, str]:
        """Return all the metadata associated with a given media."""
        resp = self._client._api(
            METADATA_GET, params={"objectId": media_id, "objectType": "media"}
        )
        _, id_to_metadata = self._client.get_metadata_mappings(self._client._project_id)
        return {id_to_metadata[int(k)]: v for k, v in resp["data"].items()}


class _SelectOption(PrettyPrintable):
    def __init__(self, selected_media: List[int]) -> None:
        self.selected_media = selected_media
        self.unselected_media: List[Union[int, List[int]]] = []
        self.field_filter_map: Dict[str, Any] = {}
        self.column_filter_map: Dict[str, Any] = {}
        self.is_unselect_mode = False


class _Project(PrettyPrintable):
    def __init__(
        self,
        project_id: int,
        dataset_id: int,
    ) -> None:
        self.project_id = project_id
        self.dataset_id = dataset_id


class _MetadataUploadRequestBody(PrettyPrintable):
    def __init__(
        self,
        selectOption: _SelectOption,
        project: _Project,
        metadata: Dict[str, Any],
    ) -> None:
        self.selectOption = selectOption
        self.project = project
        self.metadata = metadata



"""Operations (create, probe, save, etc) around images, videos from various sources (webcam, video, RTSP stream etc)."""

import io
from pathlib import Path
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np
import PIL.Image


def probe_video(video_file: str, samples_per_second: float) -> Tuple[int, int, float]:
    """Probe a video file to get some metadata before sampling images.
    Parameters
    ----------
    video_file: The local path to the video file
    samples_per_second: Number of images to sample per second
    Returns
    -------
    A tuple of three values
        - The total number of frames,
        - The number of frames to sample,
        - The video length in seconds.
    """
    if not Path(video_file).exists():
        raise FileNotFoundError(f"Video file {video_file} does not exist.")
    cap = cv2.VideoCapture(video_file)
    fps = cap.get(cv2.CAP_PROP_FPS)
    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    video_length_seconds = total_frames / fps
    sample_size = int(video_length_seconds * samples_per_second)
    cap.release()
    return (total_frames, sample_size, video_length_seconds)


def sample_images_from_video(
    video_file: str, output_dir: Path, samples_per_second: float = 1
) -> List[str]:
    """Sample images from a video file.
    Parameters
    ----------
    video_file: The local path to the video file
    output_dir: The local directory path that stores the sampled images
    samples_per_second: The number of images to sample per second. If set to zero, it disables sampling
    Returns
    -------
    a list of local file paths to the sampled images
    """
    output_dir.mkdir(parents=True, exist_ok=True)
    if samples_per_second != 0:
        total_frames, sample_size, _ = probe_video(video_file, samples_per_second)
        # Calculate the frame interval based on the desired frame rate
        sample_interval = int(total_frames / sample_size)
    else:
        sample_interval = 1
    frame_count = 0
    output = []
    cap = cv2.VideoCapture(video_file)
    while cap.isOpened():
        ret, frame = cap.read()
        # Check if end of video file
        if not ret:
            break
        # Check if the current frame should be saved
        if frame_count % sample_interval == 0:
            output_file_path = str(output_dir / f"{frame_count}.jpg")
            cv2.imwrite(output_file_path, frame)
            output.append(output_file_path)
        frame_count += 1
    cap.release()
    return output


def take_photo_from_webcam(webcam_source: Union[str, int] = 0) -> PIL.Image.Image:
    """Open a Window and allow users to take a photo from the webcam.
    It supports running on both local environment and Google's colab notebooks.
    For Google colab, it will open a window to allow users to take a photo from the webcam by pressing the "Capture" button.
    For local environment, it will open a window to allow users to take a photo from the webcam by pressing the "space" button.
    """
    try:
        # Running on Google's colab or local Jupyter notebook
        from base64 import b64decode

        from google.colab.output import eval_js  # type: ignore
        from IPython.display import Javascript, display

        # Define function to acquire images either directly from the local webcam (i.e. jupyter notebook)or from the web browser (i.e. collab)
        def take_photo() -> PIL.Image.Image:
            quality = 0.8
            js = Javascript(
                """
            async function takePhoto(quality) {
                const div = document.createElement('div');
                const capture = document.createElement('button');
                capture.textContent = 'Capture';
                div.appendChild(capture);

                const video = document.createElement('video');
                video.style.display = 'block';
                const stream = await navigator.mediaDevices.getUserMedia({video: true});

                document.body.appendChild(div);
                div.appendChild(video);
                video.srcObject = stream;
                await video.play();

                // Resize the output to fit the video element.
                google.colab.output.setIframeHeight(document.documentElement.scrollHeight, true);

                // Wait for Capture to be clicked.
                await new Promise((resolve) => capture.onclick = resolve);

                const canvas = document.createElement('canvas');
                canvas.width = video.videoWidth;
                canvas.height = video.videoHeight;
                canvas.getContext('2d').drawImage(video, 0, 0);
                stream.getVideoTracks()[0].stop();
                div.remove();
                return canvas.toDataURL('image/jpeg', quality);
            }
            """
            )
            display(js)
            data = eval_js("takePhoto({})".format(quality))
            binary = b64decode(data.split(",")[1])
            return PIL.Image.open(io.BytesIO(binary))

    except ModuleNotFoundError:
        # Capture image from local webcam using OpenCV
        import cv2

        def take_photo() -> PIL.Image.Image:
            cam = cv2.VideoCapture(webcam_source)
            cv2.namedWindow("Press space to take photo")
            cv2.startWindowThread()
            captured_frame: Optional[np.ndarray] = None
            while True:
                ret, frame = cam.read()
                if not ret:
                    print("failed to grab frame")
                    exit()
                cv2.imshow("Press space to take photo", frame)
                k = cv2.waitKey(1)
                if k % 256 == 32:
                    # SPACE pressed
                    captured_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    break
            cam.release()
            cv2.waitKey(1)
            cv2.destroyAllWindows()
            cv2.waitKey(1)
            return PIL.Image.fromarray(captured_frame)

    return take_photo()



"""Module that defines all the exceptions used in the package, and APIs for request/response error handling."""

import logging
from pprint import pformat
from typing import Any, Dict, List, Union, cast

from requests import PreparedRequest, Response
from requests.structures import CaseInsensitiveDict

_LOGGER = logging.getLogger(__name__)


class InvalidApiKeyError(Exception):
    """Exception raised when the an invalid API key is provided. This error could be raised from any SDK code, not limited to a HTTP client."""

    def __init__(self, message: str):
        self.message = f"""{message}
For more information, see https://landing-ai.github.io/landingai-python/landingai.html#manage-api-credentials"""
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class UnauthorizedError(Exception):
    """Exception raised when the user is not authorized to access the resource. Status code: 401."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class BadRequestError(Exception):
    """Exception raised when the request is not invalid. It could be due to that the Endpoint ID doesn't exist. Status code: 400, 413, 422 or other 4xx."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class PermissionDeniedError(Exception):
    """Exception raised when the requested action is not allowed. It could be due to that the user runs out of credits or the enterprise contract expired, etc. Status code: 403."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class RateLimitExceededError(Exception):
    """Exception raised when the inference request is rate limited. Status code: 429."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class ServiceUnavailableError(Exception):
    """Exception raised when the requested service is unavailable temporarily. Status code: 502, 503, 504."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class InternalServerError(Exception):
    """Exception raised when the server encounters an unexpected condition that prevents it from fulfilling the request. Status code: 5xx."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class ClientError(Exception):
    """Exception raised when the server failed due to client errors. Status code: 4xx."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class HttpError(Exception):
    """Exception raised when there's something wrong with the HTTP request. This is a generic exception that is raised when no other more specific exception is appropriate. Status code: 3xx, 4xx or 5xx."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class DuplicateUploadError(Exception):
    """Exception raised when the uploaded media is already exists in the project. Status code: 409."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class UnexpectedRedirectError(Exception):
    """Exception raised when the client encounters an unexpected redirect. Status code: 3xx."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class HttpResponse:
    """Abstraction for the HTTP response from the prediction endpoint.
    It encapulates more granular and friendly error handling logic that tailored to LandingLens.
    """

    def __init__(
        self,
        status_code: int,
        reason: str,
        uri: str,
        data: Union[str, Dict[str, Any]],
        headers: CaseInsensitiveDict,
        content_length: int,
        request: PreparedRequest,
    ) -> None:
        self.status_code = status_code
        self.reason = reason
        self.uri = uri
        self.data = data
        self.headers = headers
        self.content_length = content_length
        self.request = request

    def __str__(self) -> str:
        return pformat(self.__dict__)

    def json(self) -> Union[Dict[str, Any], List[Any]]:
        """Get json body from the response.
        NOTE: make sure you have checked the response and you are expecting a JSON body in response

        Returns
        -------
        Union[Dict[str, Any], List[Any]]
            the json body of the response. Currently, it could be either a dict or a list. (It may support more types in the future)
        """
        assert isinstance(self.data, dict) or isinstance(
            self.data, list
        ), f"expecting a dict or list instance, but it's f{type(self.data)} typed."
        return self.data

    @classmethod
    def from_response(cls, response: Response) -> "HttpResponse":
        """Create an instance of _HttpResponse from a requests.Response object."""
        try:
            content: Union[str, Dict[str, Any]] = response.json()
        except ValueError:
            # In some circumstance, the endpoint may return string content instead.
            content = str(response.content, "utf-8")
        return cls(
            status_code=response.status_code,
            reason=response.reason,
            uri=response.url,
            data=content,
            headers=cast(CaseInsensitiveDict, response.headers),
            content_length=len(response.content),
            request=response.request,
        )

    def raise_for_status(self) -> None:
        """Raises an exception if the HTTP status is erroneous.
        This method is more granular than the one provided by requests.Response.
        It's tailored to LandingLens model inference endpoints.
        """
        if self.is_2xx:
            return
        response_detail_str = f"Response detail:\n{self}"
        _LOGGER.error(
            f"Request failed with status code {self.status_code}: {self.reason}\n{response_detail_str}"
        )
        if self.status_code == 401:
            raise UnauthorizedError(
                f"Unauthorized. Please check your API key and API secret is correct.\n{response_detail_str}"
            )
        elif self.status_code == 403:
            raise PermissionDeniedError(
                f"Permission denied. Please check your account has enough credits or your enterprise contract is not expired or if you have access to this endpoint. Contact your account admin for more information.\n{response_detail_str}"
            )
        elif self.status_code == 404:
            raise BadRequestError(
                f"Endpoint doesn't exist. Please check the inference url path and other configuration is correct and try again. If this issue persists, please report this issue to LandingLens support for further assistant.\n{response_detail_str}"
            )
        elif self.status_code == 422:
            raise BadRequestError(
                f"Bad request. Please check your Endpoint ID is correct.\n{response_detail_str}"
            )
        elif self.status_code == 429:
            raise RateLimitExceededError(
                f"Rate limit exceeded. You have sent too many requests in a minute. Please wait for a minute before sending new requests. Contact your account admin or LandingLens support for how to increase your rate limit.\n{response_detail_str}"
            )
        elif self.status_code in [502, 503, 504]:
            raise ServiceUnavailableError(
                f"Service temporarily unavailable. Please try again in a few minutes.\n{response_detail_str}"
            )
        elif self.is_3xx:  # not expecting any 3xx response
            raise UnexpectedRedirectError(
                "Unexpected redirect. Please report this issue to LandingLens support for further assistant."
            )
        elif self.is_4xx:  # other 4xx errors
            raise ClientError(
                f"Client error. Please check your configuration and inference request is well-formed and try again. If this issue persists, please report this issue to LandingLens support for further assistant.\n{response_detail_str}"
            )
        elif self.is_5xx:  # other 5xx errors
            raise InternalServerError(
                f"Internal server error. The model server encountered an unexpected condition and failed. Please report this issue to LandingLens support for further assistant.\n{response_detail_str}"
            )
        raise AssertionError(
            f"Unexpected status code: {self.status_code}. Please report this issue to LandingLens support.\n{response_detail_str}"
        )

    @property
    def is_2xx(self) -> bool:
        return self.status_code < 300

    @property
    def is_3xx(self) -> bool:
        return 300 <= self.status_code < 400

    @property
    def is_4xx(self) -> bool:
        return 400 <= self.status_code < 500

    @property
    def is_5xx(self) -> bool:
        return 500 <= self.status_code < 600



import io
import json
import logging
import posixpath
from functools import lru_cache
from importlib.metadata import version
from typing import Any, Dict, Optional, Tuple, cast

import aiohttp
import requests

from landingai.data_management.utils import to_camel_case
from landingai.exceptions import HttpError
from landingai.utils import load_api_credential

METADATA_ITEMS = "metadata_items"
METADATA_UPDATE = "metadata_update"
METADATA_GET = "metadata_get"
MEDIA_LIST = "media_list"
MEDIA_REGISTER = "media_register"
MEDIA_SIGN = "media_sign"
MEDIA_DETAILS = "media_details"
MEDIA_UPDATE_SPLIT = "media_update_split"
GET_PROJECT_SPLIT = "get_project_split"
GET_PROJECT = "get_project"
GET_DEFECTS = "get_defects"
GET_PROJECT_MODEL_INFO = "get_project_model_info"
GET_FAST_TRAINING_EXPORT = "get_fast_training_export"


ROUTES = {
    GET_PROJECT_SPLIT: {
        "root_url": "LANDING_API",
        "endpoint": "api/project/split",
        "method": requests.get,
    },
    MEDIA_UPDATE_SPLIT: {
        "root_url": "LANDING_API",
        "endpoint": "api/dataset/update_media_split",
        "method": requests.post,
    },
    GET_PROJECT_MODEL_INFO: {
        "root_url": "LANDING_API",
        "endpoint": "api/registered_model/get_project_model_info",
        "method": requests.get,
    },
    GET_FAST_TRAINING_EXPORT: {
        "root_url": "LANDING_API",
        "endpoint": "api/dataset/export/fast_training_export",
        "method": requests.get,
    },
    GET_DEFECTS: {
        "root_url": "LANDING_API",
        "endpoint": "api/defect/defects",
        "method": requests.get,
    },
    METADATA_ITEMS: {
        "root_url": "LANDING_API",
        "endpoint": "api/{version}/metadata/get_metadata_by_projectId",
        "method": requests.get,
    },
    METADATA_UPDATE: {
        "root_url": "LANDING_API",
        "endpoint": "api/{version}/object/medias_metadata",
        "method": requests.post,
    },
    METADATA_GET: {
        "root_url": "LANDING_API",
        "endpoint": "api/{version}/object/metadata",
        "method": requests.get,
    },
    MEDIA_REGISTER: {
        "root_url": "LANDING_API",
        "endpoint": "api/{version}/medias/new",
        "method": requests.post,
    },
    MEDIA_SIGN: {
        "root_url": "LANDING_API",
        "endpoint": "api/{version}/medias/sign",
        "method": requests.post,
    },
    MEDIA_LIST: {
        "root_url": "LANDING_API",
        "endpoint": "api/{version}/dataset/medias",
        "method": requests.get,
    },
    MEDIA_DETAILS: {
        "root_url": "LANDING_API",
        "endpoint": "api/dataset/media_details",
        "method": requests.get,
    },
    GET_PROJECT: {
        "root_url": "LANDING_API",
        "endpoint": "api/{version}/project/with_users",
        "method": requests.get,
    },
}

_URL_ROOTS = {
    "LANDING_API": "https://app.landing.ai",
}
_API_VERSION = "v1"
_LOGGER = logging.getLogger(__name__)
_LRU_CACHE_SIZE = 1000


# Backward incompatible changes compared to LandingLens CLI:
# 1. Remove support for API key v1
# 2. Remove support for ~/.landinglens/config.ini file
# 3. Remove the no-pagination parameter in media.ls()
# 4. Rename upload() to update() in metadata.py


class LandingLens:
    """
    LandingLens client

    Example
    -------
    # Create a client by specifying API Key and project id
    >>> client = LandingLens(project, api_key)

    Parameters
    ----------
    project_id: int
        LandingLens project id.  Can override this default in individual commands.
    api_key: Optional[str]
        LandingLens API Key. If it's not provided, it will be read from the environment variable LANDINGAI_API_KEY, or from .env file on your project root directory.
    """

    def __init__(self, project_id: int, api_key: Optional[str] = None):
        self.project_id = project_id
        if not api_key:
            api_key = load_api_credential().api_key
        self.api_key = api_key

    @property
    def _project_id(self) -> int:
        return self.project_id

    @property
    def _api_key(self) -> str:
        return self.api_key

    async def _api_async(
        self,
        route_name: str,
        params: Optional[Dict[str, Any]] = None,
        resp_with_content: Optional[Dict[str, Any]] = None,
        url_replacements: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Returns a response from the LandingLens API"""
        assert resp_with_content is not None
        endpoint, headers, params, root_url, route = self._api_common_setup(
            resp_with_content, params, route_name, url_replacements
        )

        async with aiohttp.ClientSession() as session:
            # TODO: should be library agnostic
            if route["method"] == requests.get:
                method = session.get
            elif route["method"] == requests.post:
                method = session.post
            else:
                raise NotImplementedError()

            async with method(
                endpoint,
                headers=headers,
                json=resp_with_content,
                params=params,
                ssl=True,
            ) as resp:
                _LOGGER.debug("Request URL: ", resp.request_info.url)
                _LOGGER.debug("Response Code: ", resp.status)
                _LOGGER.debug("Response Reason: ", resp.reason)

                if resp.status != 200:
                    try:
                        content = await resp.text()
                        error_message = json.load(io.StringIO(content))["message"]
                    except Exception as e:
                        error_message = e
                    raise HttpError(
                        "HTTP request to LandingLens server failed with "
                        f"code {resp.status}-{resp.reason} and error message: \n"
                        f"{error_message}"
                    )
                resp_with_content = await resp.json()
                _LOGGER.debug(
                    "Response Content (500 chars): ",
                    json.dumps(resp_with_content)[:500],
                )
                assert resp_with_content is not None
                return resp_with_content

    def _api(
        self,
        route_name: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        url_replacements: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Returns a response from the LandingLens API"""
        endpoint, headers, params, root_url, route = self._api_common_setup(
            data, params, route_name, url_replacements
        )
        resp = route["method"](
            endpoint,
            params=params,
            json=data,
            headers=headers,
            verify=True,
        )
        _LOGGER.info(f"Request URL: {resp.request.url}")
        _LOGGER.debug("Response Code: ", resp.status_code)
        _LOGGER.debug("Response Reason: ", resp.reason)
        _LOGGER.debug("Response Content (500 chars): ", resp.content[:500])
        if not resp.ok:
            try:
                error_message = json.load(io.StringIO(resp.content.decode("utf-8")))[
                    "message"
                ]
            except Exception as e:
                _LOGGER.warning(f"Failed to parse error message into json: {e}")
                error_message = resp.text
            raise HttpError(
                "HTTP request to LandingLens server failed with "
                f"code {resp.status_code}-{resp.reason} and error message: \n"
                f"{error_message}"
            )
        return cast(Dict[str, Any], resp.json())

    def _api_common_setup(
        self,
        data: Optional[Dict[str, Any]],
        params: Optional[Dict[str, Any]],
        route_name: str,
        url_replacements: Optional[Dict[str, Any]],
    ) -> Tuple[str, Dict[str, Any], Dict[str, Any], str, Dict[str, Any]]:
        route = ROUTES[route_name]
        headers = {
            "apikey": self.api_key,
            "Content-Type": "application/json",
            "User-Agent": "landingai-python-" + version("landingai"),
        }
        root_url_type = cast(str, route["root_url"])

        if root_url_type not in _URL_ROOTS:
            raise ValueError(f"Unknown URL specified: {root_url_type}")

        root_url = _URL_ROOTS[root_url_type]

        if not params:
            params = {}
        if route["method"] == requests.get and not params.get("projectId"):
            params["projectId"] = self.project_id
        if route["method"] == requests.post and data and not data.get("projectId"):
            data["projectId"] = self.project_id
        endpoint = posixpath.join(root_url, cast(str, route["endpoint"]))

        if url_replacements:
            endpoint = endpoint.format(
                **{**{"version": _API_VERSION}, **url_replacements}
            )
        else:
            endpoint = endpoint.format(**{"version": _API_VERSION})

        return endpoint, headers, params, root_url, route

    def get_project_property(
        self, project_id: int, property: Optional[str] = None
    ) -> Any:
        resp = self._api(GET_PROJECT, params={"projectId": project_id})
        project = resp.get("data")
        if property is None:
            return project
        assert project is not None
        property_value = project.get(to_camel_case(property))
        if property_value is None:
            raise HttpError(f"{property} Id not found")
        return property_value

    @lru_cache(maxsize=_LRU_CACHE_SIZE)
    def get_metadata_mappings(
        self, project_id: int
    ) -> Tuple[Dict[str, Any], Dict[int, str]]:
        resp = self._api(METADATA_ITEMS, params={"projectId": project_id})
        metadata_mapping_resp = resp.get("data", {})

        metadata_mapping = {
            metadata_field["name"]: (
                metadata_field["id"],
                metadata_field["predefinedChoices"],
            )
            for metadata_field in metadata_mapping_resp.values()
        }
        id_to_metadata = {v[0]: k for k, v in metadata_mapping.items()}

        return metadata_mapping, id_to_metadata
